# Copyright 2015 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from test_root import X, cl
from EvoDAG.node import Add
from EvoDAG.base import tonparray
from SparseArray import SparseArray
import numpy as np
from nose.tools import assert_almost_equals
from test_command_line import default_nargs


def create_problem_node(nargs=4, seed=0):
    from EvoDAG import RootGP
    gp = RootGP(generations=1, popsize=nargs, classifier=False, seed=seed)
    gp.X = X
    gp.Xtest = X
    y = cl.copy()
    mask = y == 0
    y[mask] = 1
    y[~mask] = -1
    gp.y = y
    return gp, [gp.X[x] for x in range(nargs)]


def test_nargs_function():
    from EvoDAG.node import Mul
    default_nargs()
    Add.nargs = 4
    assert Add.nargs == 4
    assert Mul.nargs == 2
    default_nargs()


def test_node_pickle():
    import pickle
    import tempfile
    gp, args = create_problem_node()
    n = Add(list(range(len(args))), ytr=gp._ytr, mask=gp._mask)
    n.position = 10
    assert n.eval(args)
    with tempfile.TemporaryFile('w+b') as io:
        pickle.dump(n, io)
        io.seek(0)
        n1 = pickle.load(io)
        assert n1._mask.SSE(n._mask) == 0


def test_node_tostore():
    gp, args = create_problem_node(nargs=4)
    Add.nargs = 4
    n = Add(list(range(len(args))), ytr=gp._ytr, mask=gp._mask)
    n.position = 10
    assert n.eval(args)
    n1 = n.tostore()
    assert n1.nargs == n.nargs
    assert n1.position == n.position
    assert np.all(n1.weight == n.weight)
    assert n1.hy is None


def test_compute_weight():
    gp, args = create_problem_node()
    mask = SparseArray.fromlist(np.ones(len(gp._mask)))
    n = Add(list(range(len(args))), ytr=gp._ytr, mask=mask)
    D = np.array([tonparray(i.hy) for i in args]).T
    coef = n.compute_weight([x.hy for x in args])
    r = np.linalg.lstsq(D, tonparray(gp._ytr))[0]
    [assert_almost_equals(a, b) for a, b in zip(coef, r)]


def test_node_add():
    gp, args = create_problem_node()
    n = Add(list(range(len(args))), ytr=gp._ytr, mask=gp._mask)
    D = np.array([tonparray(i.hy) for i in args]).T
    coef = n.compute_weight([x.hy for x in args])
    assert n.eval(args)
    # a = map(lambda (a, b): a.hy * b, zip(args, coef))
    a = [_a.hy * b for _a, b in zip(args, coef)]
    r = n.cumsum(a)
    print((D * coef).sum(axis=1)[:4], "*")
    print(tonparray(n.hy)[:4])
    print(tonparray(r)[:4])
    print(tonparray(gp._mask)[:4])
    assert n.hy.SSE(r) == 0
    assert n.hy_test.SSE(r) == 0


def test_node_mul():
    from EvoDAG.node import Mul
    gp, args = create_problem_node()
    r = Mul.cumprod([x.hy for x in args])
    n = Mul(list(range(len(args))), ytr=gp._ytr, mask=gp._mask)
    coef = n.compute_weight([r])[0]
    assert n.eval(args)
    r = r * coef
    assert n.hy.SSE(r) == 0
    assert n.hy_test.SSE(r) == 0


def test_node_div():
    from EvoDAG.node import Div
    gp, args = create_problem_node(nargs=2)
    a, b = args
    r = a.hy / b.hy
    n = Div(list(range(len(args))), ytr=gp._ytr, mask=gp._mask)
    coef = n.compute_weight([r])[0]
    assert n.eval(args)
    r = r * coef
    assert n.hy.SSE(r) == 0
    assert n.hy_test.SSE(r) == 0


def test_node_fabs():
    from EvoDAG.node import Fabs
    gp, args = create_problem_node(nargs=1)
    r = args[0].hy.fabs()
    n = Fabs(list(range(len(args))), ytr=gp._ytr, mask=gp._mask)
    coef = n.compute_weight([r])[0]
    assert n.eval(args)
    r = r * coef
    assert n.hy.SSE(r) == 0
    assert n.hy_test.SSE(r) == 0


def test_node_exp():
    from EvoDAG.node import Exp
    gp, args = create_problem_node(nargs=1)
    r = args[0].hy.exp()
    n = Exp(list(range(len(args))), ytr=gp._ytr, mask=gp._mask)
    coef = n.compute_weight([r])[0]
    assert n.eval(args)
    r = r * coef
    assert n.hy.SSE(r) == 0
    assert n.hy_test.SSE(r) == 0


def test_node_sqrt():
    from EvoDAG.node import Sqrt
    gp, args = create_problem_node(nargs=1)
    r = args[0].hy.sqrt()
    n = Sqrt(list(range(len(args))), ytr=gp._ytr, mask=gp._mask)
    coef = n.compute_weight([r])[0]
    assert n.eval(args)
    r = r * coef
    assert n.hy.SSE(r) == 0
    assert n.hy_test.SSE(r) == 0


def test_node_sin():
    from EvoDAG.node import Sin
    gp, args = create_problem_node(nargs=1)
    r = args[0].hy.sin()
    n = Sin(list(range(len(args))), ytr=gp._ytr, mask=gp._mask)
    coef = n.compute_weight([r])[0]
    assert n.eval(args)
    r = r * coef
    assert n.hy.SSE(r) == 0
    assert n.hy_test.SSE(r) == 0


def test_node_cos():
    from EvoDAG.node import Cos
    gp, args = create_problem_node(nargs=1)
    r = args[0].hy.cos()
    n = Cos(list(range(len(args))), ytr=gp._ytr, mask=gp._mask)
    coef = n.compute_weight([r])[0]
    assert n.eval(args)
    r = r * coef
    assert n.hy.SSE(r) == 0
    assert n.hy_test.SSE(r) == 0


def test_node_ln():
    from EvoDAG.node import Log1p
    gp, args = create_problem_node(nargs=1)
    r = args[0].hy.log1p()
    n = Log1p(list(range(len(args))), ytr=gp._ytr, mask=gp._mask)
    coef = n.compute_weight([r])[0]
    assert n.eval(args)
    r = r * coef
    assert n.hy.SSE(r) == 0
    assert n.hy_test.SSE(r) == 0


def test_node_sq():
    from EvoDAG.node import Sq
    gp, args = create_problem_node(nargs=1)
    r = args[0].hy.sq()
    n = Sq(list(range(len(args))), ytr=gp._ytr, mask=gp._mask)
    coef = n.compute_weight([r])[0]
    assert n.eval(args)
    r = r * coef
    assert n.hy.SSE(r) == 0
    assert n.hy_test.SSE(r) == 0


def test_node_min():
    from EvoDAG.node import Min
    gp, args = create_problem_node(nargs=3)
    r = args[0].hy.min(args[1].hy).min(args[2].hy)
    n = Min(list(range(len(args))), ytr=gp._ytr, mask=gp._mask)
    coef = n.compute_weight([r])
    assert n.eval(args)
    r = r * coef
    assert n.hy.SSE(r) == 0
    assert n.hy_test.SSE(r) == 0
    assert n.weight == coef


def test_node_max():
    from EvoDAG.node import Max
    gp, args = create_problem_node(nargs=3)
    r = args[0].hy.max(args[1].hy).max(args[2].hy)
    n = Max(list(range(len(args))), ytr=gp._ytr, mask=gp._mask)
    coef = n.compute_weight([r])
    assert n.eval(args)
    r = r * coef
    assert n.hy.SSE(r) == 0
    assert n.hy_test.SSE(r) == 0
    assert n.weight == coef


def test_node_symbol():
    from EvoDAG.node import Add, Mul, Div, Fabs,\
        Exp, Sqrt, Sin, Cos, Log1p,\
        Sq, Min, Max
    for f, s in zip([Add, Mul, Div, Fabs,
                     Exp, Sqrt, Sin, Cos, Log1p,
                     Sq, Min, Max],
                    ['+', '*', '/', 'fabs', 'exp',
                     'sqrt', 'sin', 'cos', 'log1p',
                     'sq', 'min', 'max']):
        assert f.symbol == s


def test_node_hash():
    from EvoDAG.node import Add, Mul
    add = Add([21, 3])
    sets = set()
    sets.add(add.signature())
    add2 = Add([21, 3])
    assert add2.signature() in sets
    assert Add([3, 21]).signature() in sets
    assert Mul([2, 3]).signature() == Mul([3, 2]).signature()


def test_variable_multiple_output():
    from EvoDAG.node import Variable
    gp, args = create_problem_node(nargs=4, seed=0)
    gp2, _ = create_problem_node(nargs=4, seed=1)
    n1 = Variable(0, ytr=gp._ytr, mask=gp._mask)
    n1.eval(args)
    print(n1.weight)
    n = Variable(0, ytr=[gp._ytr, gp._ytr],
                 mask=[gp._mask, gp2._mask])
    n.eval(args)
    assert n.weight[0] != n.weight[1]


def test_variable_multiple_output_isfinite():
    from EvoDAG.node import Variable
    gp, args = create_problem_node(nargs=4, seed=0)
    gp2, _ = create_problem_node(nargs=4, seed=1)
    n1 = Variable(0, ytr=gp._ytr, mask=gp._mask)
    n1.eval(args)
    print(n1.weight)
    n = Variable(0, ytr=[gp._ytr, gp._ytr],
                 mask=[gp._mask, gp2._mask])
    n.eval(args)
    n.isfinite()


def test_Add_multiple_output():
    from EvoDAG.node import Variable
    gp, args = create_problem_node(nargs=4, seed=0)
    gp2, _ = create_problem_node(nargs=4, seed=1)
    ytr = [gp._ytr, gp._ytr]
    mask = [gp._mask, gp2._mask]
    vars = [Variable(k, ytr=ytr, mask=mask) for k in range(len(args))]
    [x.eval(args) for x in vars]
    add = Add(range(len(vars)), ytr=ytr, mask=mask)
    assert add.eval(vars)
    gp, args = create_problem_node(nargs=4, seed=0)
    vars = [Variable(k, ytr=gp._ytr, mask=gp._mask) for k in range(len(args))]
    [x.eval(args) for x in vars]
    add2 = Add(range(len(vars)), ytr=gp._ytr, mask=gp._mask)
    assert add2.eval(vars)
    assert add2.hy.SSE(add.hy[0]) == 0
    assert add2.hy_test.SSE(add.hy_test[0]) == 0


def test_Add_multiple_output2():
    from EvoDAG.node import Variable
    gp, args = create_problem_node(nargs=4, seed=0)
    for i in args:
        i.hy_test = None
    gp2, _ = create_problem_node(nargs=4, seed=1)
    ytr = [gp._ytr, gp._ytr]
    mask = [gp._mask, gp2._mask]
    vars = [Variable(k, ytr=ytr, mask=mask) for k in range(len(args))]
    [x.eval(args) for x in vars]
    add = Add(range(len(vars)), ytr=ytr, mask=mask)
    assert add.eval(vars)
    vars = [Variable(k, ytr=gp._ytr, mask=gp._mask) for k in range(len(args))]
    [x.eval(args) for x in vars]
    add2 = Add(range(len(vars)), ytr=gp._ytr, mask=gp._mask)
    assert add2.eval(vars)
    assert add2.hy.SSE(add.hy[0]) == 0
    assert isinstance(add.weight, list) and len(add.weight) == 2
    assert isinstance(add2.weight, np.ndarray)


def test_one_multiple_output():
    from EvoDAG.node import Variable
    from EvoDAG.node import Fabs, Exp, Sqrt, Sin, Cos, Log1p, Sq
    from EvoDAG.node import Acos, Asin, Atan, Tan, Cosh, Sinh, Tanh
    from EvoDAG.node import Acosh, Asinh, Atanh, Expm1, Log, Log2, Log10, Lgamma
    from EvoDAG.node import Sign, Ceil, Floor

    for flag in [False, True]:
        gp, args = create_problem_node(nargs=4, seed=0)
        if flag:
            for i in args:
                i.hy_test = None
        gp2, _ = create_problem_node(nargs=4, seed=1)
        ytr = [gp._ytr, gp._ytr]
        mask = [gp._mask, gp2._mask]
        vars = [Variable(k, ytr=ytr, mask=mask) for k in range(len(args))]
        [x.eval(args) for x in vars]
        vars2 = [Variable(k, ytr=gp._ytr, mask=gp._mask) for k in range(len(args))]
        [x.eval(args) for x in vars2]
        for FF in [Fabs, Exp, Sqrt, Sin, Cos, Log1p, Sq,
                   Acos, Asin, Atan, Tan, Cosh, Sinh, Tanh,
                   Acosh, Asinh, Atanh, Expm1, Log, Log2, Log10, Lgamma,
                   Sign, Ceil, Floor]:
            ff = FF(0, ytr=ytr, mask=mask)
            ff.eval(vars)
            ff2 = FF(0, ytr=gp._ytr, mask=gp._mask)
            ff2.eval(vars2)
            print(ff.hy)
            if ff.hy is None:
                continue
            assert isinstance(ff.weight, list) and len(ff.weight) == 2
            assert isinstance(ff2.weight, float)
            hy = ff.hy[0]
            if hy.isfinite():
                print('*', FF)
                assert hy.SSE(ff2.hy) == 0


def test_functions_w_multiple_output():
    from EvoDAG.node import Variable, Mul, Div, Min, Max, Atan2, Hypot
    from EvoDAG.node import Argmax, Argmin
    for ff in [Mul, Div, Min, Max, Atan2, Hypot, Argmax, Argmin]:
        for flag in [False, True]:
            gp, args = create_problem_node(nargs=4, seed=0)
            if flag:
                for i in args:
                    i.hy_test = None
            gp2, _ = create_problem_node(nargs=4, seed=1)
            ytr = [gp._ytr, gp._ytr]
            mask = [gp._mask, gp2._mask]
            vars = [Variable(k, ytr=ytr, mask=mask) for k in range(len(args))]
            [x.eval(args) for x in vars]
            mul = ff(range(len(vars)), ytr=ytr, mask=mask)
            assert mul.eval(vars)
            gp, args = create_problem_node(nargs=4, seed=0)
            vars = [Variable(k, ytr=gp._ytr, mask=gp._mask) for k in range(len(args))]
            [x.eval(args) for x in vars]
            mul2 = ff(range(len(vars)), ytr=gp._ytr, mask=gp._mask)
            assert mul2.eval(vars)
            assert mul2.hy.SSE(mul.hy[0]) == 0
            if mul.hy_test is not None:
                assert mul2.hy_test.SSE(mul.hy_test[0]) == 0
            

def test_density_safe():
    gp, args = create_problem_node(nargs=4, seed=0)
    for i in gp._function_set:
        assert i.density_safe or not i.density_safe


def test_functions_finite():
    from EvoDAG.node import Variable, Mul, Div, Min, Max, Atan2, Hypot
    from EvoDAG.node import Argmax, Argmin
    for ff in [Mul, Div, Min, Max, Atan2, Hypot, Argmax, Argmin]:
        for flag in [False, True]:
            gp, args = create_problem_node(nargs=4, seed=0)
            for v in args:
                _ = [x for x in v._eval_tr.full_array()]
                _[0] = float('inf')
                _[1] = float('inf')
                _[3] = 0
                _[4] = 0
                v._eval_tr = SparseArray.fromlist(_)
            if flag:
                for i in args:
                    i.hy_test = None
            gp2, _ = create_problem_node(nargs=4, seed=1)
            # ytr = [gp._ytr, gp._ytr]
            ytr = gp._ytr
            # mask = [gp._mask, gp2._mask]
            mask = gp._mask.full_array()
            mask[0] = 1
            mask = SparseArray.fromlist(mask)
            vars = [Variable(k, ytr=ytr, mask=mask, finite=flag)
                    for k in range(len(args))]
            for x in vars:
                _ = x.eval(args)
                print(mask.full_array())
                assert _ == flag
            if not flag:
                continue
            mul = ff(range(len(vars)), ytr=ytr, mask=mask, finite=False)
            _ = mul.eval(vars)
            if isinstance(mul, Div):
                assert not _
                mul = ff(range(len(vars)), ytr=ytr, mask=mask, finite=True)
                _ = mul.eval(vars)
                assert _


def test_classification_regression_klass():
    from EvoDAG.node import Variable
    assert Variable.classification
    assert Variable.regression


def create_problem_node2(nargs=4, seed=0):
    from EvoDAG import RootGP
    gp = RootGP(generations=1, popsize=nargs,
                multiple_outputs=True, seed=seed)
    gp.X = X
    gp.Xtest = X
    y = cl.copy()
    gp.nclasses(y)
    gp.y = y
    return gp, [gp.X[x] for x in range(nargs)]


def test_functions_extra_args():
    from EvoDAG.node import Variable, Mul, Div, Min, Max, Atan2, Hypot
    from EvoDAG.node import Argmax, Argmin
    for ff in [Mul, Div, Min, Max, Atan2, Hypot, Argmax, Argmin]:
        gp, args = create_problem_node2(nargs=4, seed=0)
        vars = [Variable(k, ytr=gp._ytr, y_klass=gp._y_klass,
                         mask=gp._mask, finite=True)
                for k in range(len(args))]
        [x.eval(args) for x in vars]
        mul = ff(range(len(vars)), ytr=gp._ytr, klass=gp._y_klass,
                 mask=gp._mask, finite=True)
        mul.eval(vars)


def test_naive_bayes():
    import numpy as np
    from EvoDAG.node import Variable, NaiveBayes
    gp, args = create_problem_node2(nargs=4, seed=0)
    gp.random_leaf()
    vars = [Variable(k, ytr=gp._ytr, y_klass=gp._y_klass,
                     mask=gp._mask, finite=True)
            for k in range(len(args))]
    [x.eval(args) for x in vars]
    naive_bayes = NaiveBayes(range(len(vars)), ytr=gp._ytr, naive_bayes=gp._naive_bayes,
                             mask=gp._mask, finite=True)
    naive_bayes.eval(vars)
    mask = np.array(gp._mask_ts.sign().full_array(), dtype=np.bool)
    klass = np.array(gp._y_klass.full_array())[mask]
    unique_klass = np.unique(klass)
    mean = []
    std2 = []
    p_klass = []
    l = []
    for v in vars:
        l += v.hy
    for v in l:
        var = np.array(v.full_array())[mask]
        mean.append([np.mean(var[k == klass]) for k in unique_klass])
        std2.append([np.var(var[k == klass]) for k in unique_klass])
        p_klass = [(k == klass).mean() for k in unique_klass]
    mean = np.array(mean)
    std2 = np.array(std2)
    p_klass = np.array(p_klass)
    likelihood = []
    for i in range(unique_klass.shape[0]):
        a = np.log(p_klass[i])
        b = - 0.5 * np.sum([np.log(2. * np.pi * s[i]) for s in std2])
        _ = [(x + -m[i]).sq() * (1 / s[i]) for x, m, s in zip(l, mean, std2)]
        _ = SparseArray.cumsum(_) * -0.5
        likelihood.append(_ + b + a)
    for a, b in zip(likelihood, naive_bayes.hy):
        [assert_almost_equals(v, w) for v, w in zip(a, b.data)]


def test_naive_bayes_sklearn():
    from EvoDAG.naive_bayes import NaiveBayes as NB
    from EvoDAG.node import NaiveBayes
    try:
        from sklearn.naive_bayes import GaussianNB
    except ImportError:
        return

    class Var(object):
        def __init__(self, a):
            self.hy = a
            self.hy_test = None

    m = GaussianNB().fit(X, cl)
    hy = m._joint_log_likelihood(X)
    # hy = m.predict_proba(X) * 2 - 1
    vars = [Var(SparseArray.fromlist(x)) for x in X.T]
    nb = NB(mask=SparseArray.fromlist([1 for _ in X[:, 0]]), klass=SparseArray.fromlist(cl),
            nclass=3)
    naive_bayes = NaiveBayes([x for x in range(4)], naive_bayes=nb)
    naive_bayes.eval(vars)
    for a, b in zip(naive_bayes.hy, hy.T):
        for a1, b1 in zip(a.full_array(), b):
            assert_almost_equals(a1, b1, 3)


def test_naive_bayes_MN():
    import numpy as np
    from EvoDAG.node import Variable, NaiveBayesMN
    from EvoDAG.naive_bayes import NaiveBayes as MN
    gp, args = create_problem_node2(nargs=3, seed=0)
    gp.random_leaf()
    vars = [Variable(k, ytr=gp._ytr, y_klass=gp._y_klass,
                     mask=gp._mask, finite=True)
            for k in range(len(args))]
    [x.eval(args) for x in vars]
    nb = MN(mask=gp._mask_ts, klass=gp._y_klass,
            nclass=gp._labels.shape[0])
    l = []
    [[l.append(y) for y in x.hy] for x in vars]
    coef = [nb.coef_MN(x) for x in l]
    p_klass = coef[0][1]
    coef = [(k, x[0]) for k, x in enumerate(coef) if np.all(np.isfinite(np.array(x[0])))]
    R = []
    for k, p in enumerate(p_klass):
        r = p
        for v, w in coef:
            r += (l[v] * w[k])
        R.append(r)
    naive_bayes = NaiveBayesMN(range(len(vars)), ytr=gp._ytr, naive_bayes=gp._naive_bayes,
                               mask=gp._mask, finite=True)
    naive_bayes.eval(vars)
    for a, b in zip(R, naive_bayes.hy):
        [assert_almost_equals(np.exp(v), w) for v, w in zip(a.data, b.data)]


def test_naive_bayes_MN_variable():
    from EvoDAG.node import NaiveBayesMN
    gp, args = create_problem_node2(nargs=3, seed=0)
    gp.random_leaf()
    naive_bayes = NaiveBayesMN(range(len(gp.X)), ytr=gp._ytr, naive_bayes=gp._naive_bayes,
                               mask=gp._mask, finite=True)
    naive_bayes.eval(gp.X)
    assert len(naive_bayes.hy) == 3


def test_multiple_variables():
    from EvoDAG.node import MultipleVariables
    gp, args = create_problem_node2(nargs=3, seed=0)
    gp.random_leaf()
    mv = MultipleVariables([x for x in range(len(gp.X))],
                           ytr=gp._ytr, naive_bayes=gp._naive_bayes,
                           mask=gp._mask, finite=True)
    mv.eval(gp.X)
    assert len(mv.hy) == 3
    mv2 = MultipleVariables([x for x in range(len(gp.X))],
                            ytr=gp._ytr[0], mask=gp._mask[0], finite=True)
    mv2.eval(gp.X)
    assert isinstance(mv2.hy, SparseArray)
    assert mv2.hy.SSE(mv.hy[0]) == 0
    l = []
    for a, b in zip(gp.X, mv2.weight):
        l.append(a.hy * b)
    r = SparseArray.cumsum(l)
    assert mv2.hy.SSE(r) == 0


def test_centroid_variable():
    from EvoDAG.node import Centroid
    gp, args = create_problem_node2(nargs=3, seed=0)
    gp.random_leaf()
    X = np.array([x.hy.full_array() for x in gp.X]).T
    index = np.array(gp.naive_bayes._mask)
    cl = np.array(gp._y_klass.full_array(), dtype=np.int)[index]
    Xt = X[index]
    coef = [-np.mean(Xt[x == cl], axis=0) for x in range(3)]
    naive_bayes = gp.naive_bayes
    centroid = Centroid(range(len(gp.X)), ytr=gp._ytr,
                        naive_bayes=naive_bayes, finite=True)
    centroid.height = 1
    centroid.eval(gp.X)
    [[assert_almost_equals(_v, _w) for _v, _w in zip(v, w)] for v, w
     in zip(coef, centroid.weight)]
    r = [- ((X + w)**2).sum(axis=1) for w in coef]
    [[assert_almost_equals(_v, _w) for _v, _w in zip(v, w.full_array())] for v, w
     in zip(r, centroid.hy)]


def test_orthogonal_selection():
    from EvoDAG.node import Function, Add, NaiveBayes, NaiveBayesMN, Centroid
    assert Function.orthogonal_selection is False
    assert Add.orthogonal_selection
    assert NaiveBayes.orthogonal_selection
    assert NaiveBayesMN.orthogonal_selection
    assert Centroid.orthogonal_selection


def test_centroid_json():
    from EvoDAG.node import Centroid
    gp, args = create_problem_node2(nargs=3, seed=0)
    gp.random_leaf()
    naive_bayes = gp.naive_bayes
    centroid = Centroid(range(len(gp.X)), ytr=gp._ytr,
                        naive_bayes=naive_bayes, finite=True)
    centroid.height = 1
    centroid.eval(gp.X)
    assert isinstance(centroid.tojson(), str)



# Copyright 2016 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from test_root import X, cl
import numpy as np
from SparseArray import SparseArray
from nose.tools import assert_almost_equals


def test_transform_to_mo():
    from EvoDAG import EvoDAG
    y = cl.copy()
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                early_stopping_rounds=100,
                time_limit=0.9,
                multiple_outputs=True,
                seed=0,
                popsize=10000)
    gp.nclasses(y)
    k = np.unique(y)
    y = gp._bagging_fitness.transform_to_mo(y)
    assert k.shape[0] == y.shape[1]


def test_multiple_outputs():
    from EvoDAG import EvoDAG
    y = cl.copy()
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                early_stopping_rounds=100,
                time_limit=0.9,
                multiple_outputs=True,
                seed=0,
                popsize=10000)
    gp.X = X
    gp.nclasses(y)
    gp.y = y
    gp.create_population()
    assert len(gp.y) == 3


def test_multiple_outputs2():
    from EvoDAG import EvoDAG
    from EvoDAG.model import Model
    y = cl.copy()
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                early_stopping_rounds=100,
                time_limit=0.9,
                multiple_outputs=True,
                seed=0,
                popsize=10000).fit(X, y, test_set=X)
    m = gp.model()
    assert isinstance(m, Model)
    assert len(gp.y) == 3


def test_multiple_outputs_mask():
    from EvoDAG import EvoDAG
    from EvoDAG.node import Add, Min, Max
    y = cl.copy()
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                function_set=[Add, Min, Max],
                early_stopping_rounds=100,
                time_limit=0.9,
                tr_fraction=0.8,
                multiple_outputs=True,
                seed=0,
                popsize=100)
    gp.X = X[:-1]
    gp.nclasses(y[:-1])
    gp.y = y[:-1]
    assert gp._mask_vs.sum() == 27


def test_multiple_outputs_BER_vs():
    from EvoDAG import EvoDAG
    from EvoDAG.node import Add, Min, Max
    from EvoDAG.utils import BER
    y = cl.copy()
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                function_set=[Add, Min, Max],
                early_stopping_rounds=100,
                time_limit=0.9,
                multiple_outputs=True,
                seed=0,
                popsize=100)
    gp.X = X[:-1]
    gp.nclasses(y[:-1])
    gp.y = y[:-1]
    gp.create_population()
    a = gp.random_offspring()
    hy = np.array(SparseArray.argmax(a.hy).full_array())
    mask = np.array(gp._mask_vs.full_array()).astype(np.bool)
    assert_almost_equals(-a.fitness_vs * 100, BER(y[:-1][mask], hy[mask]))


def test_multiple_outputs_BER_ts():
    from EvoDAG import EvoDAG
    from EvoDAG.node import Add, Min, Max
    from EvoDAG.utils import BER
    y = cl.copy()
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                function_set=[Add, Min, Max],
                early_stopping_rounds=100,
                time_limit=0.9,
                multiple_outputs=True,
                seed=0,
                popsize=100)
    gp.X = X[:-1]
    gp.nclasses(y[:-1])
    gp.y = y[:-1]
    gp.create_population()
    a = gp.random_offspring()
    hys = SparseArray.argmax(a.hy)
    hy = np.array(hys.full_array())
    print(((hys - gp._y_klass).sign().fabs() * gp._mask_ts).sum())
    mask = np.array(gp._mask_ts.full_array()).astype(np.bool)
    assert_almost_equals(-a.fitness * 100, BER(y[:-1][mask], hy[mask]))


def test_multiple_outputs_error_rate_ts():
    from EvoDAG import EvoDAG
    from EvoDAG.node import Add, Min, Max
    y = cl.copy()
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                function_set=[Add, Min, Max],
                early_stopping_rounds=100,
                time_limit=0.9,
                multiple_outputs=True,
                fitness_function='ER',
                seed=0,
                popsize=100)
    gp.X = X[:-1]
    gp.nclasses(y[:-1])
    gp.y = y[:-1]
    gp.create_population()
    a = gp.random_offspring()
    hys = SparseArray.argmax(a.hy)
    hy = np.array(hys.full_array())
    # print(((hys - gp._y_klass).sign().fabs() * gp._mask_ts).sum())
    mask = np.array(gp._mask_ts.full_array()).astype(np.bool)
    # print((y[:-1][mask] != hy[mask]).mean())
    print(-a.fitness, (y[:-1][mask] != hy[mask]).mean())
    assert_almost_equals(-a.fitness, (y[:-1][mask] != hy[mask]).mean())


def test_multiple_outputs_ER_vs():
    from EvoDAG import EvoDAG
    from EvoDAG.node import Add, Min, Max
    y = cl.copy()
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                function_set=[Add, Min, Max],
                early_stopping_rounds=100,
                fitness_function='ER',
                time_limit=0.9,
                multiple_outputs=True,
                seed=0,
                popsize=100)
    gp.X = X[:-1]
    gp.nclasses(y[:-1])
    gp.y = y[:-1]
    gp.create_population()
    a = gp.random_offspring()
    hy = np.array(SparseArray.argmax(a.hy).full_array())
    # mask = np.array(gp._mask_vs.full_array()).astype(np.bool)
    mask = np.array(gp._mask_ts.full_array()) == 0
    print(mask)
    assert_almost_equals(-a.fitness_vs, (y[:-1][mask] != hy[mask]).mean())


def test_macro_F1():
    from EvoDAG.cython_utils import Score
    from EvoDAG import EvoDAG
    y = cl.copy()
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                early_stopping_rounds=100,
                time_limit=0.9,
                multiple_outputs=True,
                seed=2,
                popsize=1000)
    gp.y = y
    gp.X = X
    gp.create_population()
    off = gp.random_offspring()
    hy = SparseArray.argmax(off.hy)
    index = np.array(gp._mask_ts.index)
    y = np.array(gp._y_klass.full_array())[index]
    hy = np.array(hy.full_array())[index]
    nclasses = gp._bagging_fitness.nclasses
    precision = np.array([(y[hy == k] == k).mean() for k in range(nclasses)])
    recall = np.array([(hy[y == k] == k).mean() for k in range(nclasses)])
    print(precision, recall)
    f1 = Score(nclasses)
    mf1, mf1_v = f1.a_F1(gp._y_klass, SparseArray.argmax(off.hy), gp._mask_ts.index)
    for x, y in zip(precision, f1.precision):
        if not np.isfinite(x):
            continue
        assert_almost_equals(x, y)
    for x, y in zip(recall, f1.recall):
        if not np.isfinite(x):
            continue
        assert_almost_equals(x, y)
    _ = (2 * precision * recall) / (precision + recall)
    m = ~ np.isfinite(_)
    _[m] = 0
    assert_almost_equals(np.mean(_), mf1)
    print(f1.precision, f1.recall, mf1, mf1_v)
    gp._fitness_function = 'macro-F1'
    gp._bagging_fitness.set_fitness(off)
    assert_almost_equals(off.fitness, mf1 - 1)
    assert_almost_equals(off.fitness_vs, mf1_v - 1)
    index = np.array(gp._mask_ts.full_array()) == 0
    y = np.array(gp._y_klass.full_array())[index]
    hy = SparseArray.argmax(off.hy)
    hy = np.array(hy.full_array())[index]
    precision = np.array([(y[hy == k] == k).mean() for k in range(nclasses)])
    recall = np.array([(hy[y == k] == k).mean() for k in range(nclasses)])
    _ = (2 * precision * recall) / (precision + recall)
    m = ~ np.isfinite(_)
    _[m] = 0
    assert_almost_equals(np.mean(_) - 1, off.fitness_vs)


def test_F1():
    from EvoDAG.cython_utils import Score
    from EvoDAG import EvoDAG
    y = cl.copy()
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                early_stopping_rounds=100,
                time_limit=0.9,
                multiple_outputs=True,
                seed=0,
                popsize=500)
    gp.y = y
    gp.X = X
    gp.create_population()
    off = gp.random_offspring()
    hy = SparseArray.argmax(off.hy)
    index = np.array(gp._mask_ts.index)
    y = np.array(gp._y_klass.full_array())[index]
    hy = np.array(hy.full_array())[index]
    nclasses = gp._bagging_fitness.nclasses
    precision = np.array([(y[hy == k] == k).mean() for k in range(nclasses)])
    recall = np.array([(hy[y == k] == k).mean() for k in range(nclasses)])
    f1 = Score(nclasses)
    assert gp._bagging_fitness.min_class >= 0 and gp._bagging_fitness.min_class < gp._bagging_fitness.nclasses
    mf1, mf1_v = f1.F1(gp._bagging_fitness.min_class,
                       gp._y_klass, SparseArray.argmax(off.hy),
                       gp._mask_ts.index)

    _ = (2 * precision * recall) / (precision + recall)
    m = ~ np.isfinite(_)
    _[m] = 0
    assert_almost_equals(_[gp._bagging_fitness.min_class], mf1)
    gp._fitness_function = 'F1'
    gp._bagging_fitness.set_fitness(off)
    assert_almost_equals(mf1 - 1, off.fitness)
    index = np.array(gp._mask_ts.full_array()) == 0
    y = np.array(gp._y_klass.full_array())[index]
    hy = SparseArray.argmax(off.hy)
    hy = np.array(hy.full_array())[index]
    precision = np.array([(y[hy == k] == k).mean() for k in range(nclasses)])
    recall = np.array([(hy[y == k] == k).mean() for k in range(nclasses)])
    _ = (2 * precision * recall) / (precision + recall)
    m = ~ np.isfinite(_)
    _[m] = 0
    assert_almost_equals(_[gp._bagging_fitness.min_class] - 1, off.fitness_vs)


def test_min_class():
    from EvoDAG import EvoDAG
    y = cl.copy()
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                early_stopping_rounds=100,
                time_limit=0.9,
                multiple_outputs=True,
                seed=0,
                popsize=100)
    gp.y = y[:-1]
    gp.X = X[:-1]
    assert gp._bagging_fitness.min_class == 2


def test_mask():
    from EvoDAG import EvoDAG
    y = cl.copy()
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                early_stopping_rounds=100,
                time_limit=0.9,
                multiple_outputs=True,
                seed=0,
                popsize=100)
    gp.y = y
    gp.X = X
    gp.create_population()
    ts = np.array(gp._mask_ts.full_array()) == 0
    vs = np.array(gp._mask_vs.full_array()) == 1
    assert np.all(ts == vs)


def test_RecallF1():
    from EvoDAG.cython_utils import Score
    from EvoDAG import EvoDAG
    y = cl.copy()
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                early_stopping_rounds=100,
                time_limit=0.9,
                multiple_outputs=True,
                seed=0,
                popsize=500)
    gp.y = y
    gp.X = X
    gp.create_population()
    off = gp.random_offspring()
    hy = SparseArray.argmax(off.hy)
    index = np.array(gp._mask_ts.index)
    y = np.array(gp._y_klass.full_array())[index]
    hy = np.array(hy.full_array())[index]
    nclasses = gp._bagging_fitness.nclasses
    precision = np.array([(y[hy == k] == k).mean() for k in range(nclasses)])
    recall = np.array([(hy[y == k] == k).mean() for k in range(nclasses)])
    f1 = Score(nclasses)
    mf1, mf1_v = f1.macroRecallF1(gp._y_klass, SparseArray.argmax(off.hy),
                                  gp._mask_ts.index)
    assert_almost_equals(np.mean(recall), mf1)
    gp._fitness_function = 'macro-RecallF1'
    gp._bagging_fitness.set_fitness(off)
    assert_almost_equals(mf1 - 1, off.fitness)
    index = np.array(gp._mask_ts.full_array()) == 0
    y = np.array(gp._y_klass.full_array())[index]
    hy = SparseArray.argmax(off.hy)
    hy = np.array(hy.full_array())[index]
    precision = np.array([(y[hy == k] == k).mean() for k in range(nclasses)])
    recall = np.array([(hy[y == k] == k).mean() for k in range(nclasses)])
    _ = (2 * precision * recall) / (precision + recall)
    m = ~ np.isfinite(_)
    _[m] = 0
    print(precision, recall)
    print(f1.precision2, f1.recall2)
    for x, y in zip(precision, f1.precision2):
        if not np.isfinite(x):
            continue
        assert_almost_equals(x, y)
    for x, y in zip(recall, f1.recall2):
        assert_almost_equals(x, y)


def test_AccdDotMacroF1():
    from EvoDAG.cython_utils import Score
    from EvoDAG import EvoDAG
    y = cl.copy()
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                early_stopping_rounds=100,
                time_limit=0.9,
                multiple_outputs=True,
                seed=0,
                popsize=500)
    gp.y = y
    gp.X = X
    gp.create_population()
    off = gp.random_offspring()
    hy = SparseArray.argmax(off.hy)
    index = np.array(gp._mask_ts.index)
    y = np.array(gp._y_klass.full_array())[index]
    hy = np.array(hy.full_array())[index]
    nclasses = gp._bagging_fitness.nclasses
    precision = np.array([(y[hy == k] == k).mean() for k in range(nclasses)])
    recall = np.array([(hy[y == k] == k).mean() for k in range(nclasses)])
    f1 = Score(nclasses)
    mf1, mf1_v = f1.accDotMacroF1(gp._y_klass, SparseArray.argmax(off.hy),
                                  gp._mask_ts.index)
    _ = (2 * precision * recall) / (precision + recall)
    m = ~ np.isfinite(_)
    _[m] = 0
    assert_almost_equals(np.mean(_) * (y == hy).mean(), mf1)
    gp._fitness_function = 'accDotMacroF1'
    gp._bagging_fitness.set_fitness(off)
    assert_almost_equals(off.fitness, mf1 - 1)
    assert_almost_equals(off.fitness_vs, mf1_v - 1)
    index = np.array(gp._mask_ts.full_array()) == 0
    y = np.array(gp._y_klass.full_array())[index]
    hy = SparseArray.argmax(off.hy)
    hy = np.array(hy.full_array())[index]
    precision = np.array([(y[hy == k] == k).mean() for k in range(nclasses)])
    recall = np.array([(hy[y == k] == k).mean() for k in range(nclasses)])
    _ = (2 * precision * recall) / (precision + recall)
    m = ~ np.isfinite(_)
    _[m] = 0
    assert_almost_equals(np.mean(_) * (y == hy).mean() - 1, off.fitness_vs)


def test_a_precision():
    from EvoDAG.cython_utils import Score
    from EvoDAG import EvoDAG
    y = cl.copy()
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                early_stopping_rounds=100,
                time_limit=0.9,
                multiple_outputs=True,
                seed=0,
                popsize=500)
    gp.y = y
    gp.X = X
    gp.create_population()
    # off = gp.random_offspring()
    off = gp.population.bsf
    hy = SparseArray.argmax(off.hy)
    index = np.array(gp._mask_ts.index)
    y = np.array(gp._y_klass.full_array())[index]
    hy = np.array(hy.full_array())[index]
    nclasses = gp._bagging_fitness.nclasses
    precision = np.array([(y[hy == k] == k).mean() for k in range(nclasses)])
    f1 = Score(nclasses)
    mf1, mf1_v = f1.a_precision(gp._y_klass, SparseArray.argmax(off.hy),
                                gp._mask_ts.index)
    assert_almost_equals(np.mean(precision), mf1)
    gp._fitness_function = 'a_precision'
    gp._bagging_fitness.set_fitness(off)
    assert_almost_equals(mf1 - 1, off.fitness)
    index = np.array(gp._mask_ts.full_array()) == 0
    y = np.array(gp._y_klass.full_array())[index]
    hy = SparseArray.argmax(off.hy)
    hy = np.array(hy.full_array())[index]
    precision = np.array([(y[hy == k] == k).mean() for k in range(nclasses)])
    assert_almost_equals(np.mean(precision) - 1, off.fitness_vs)


def test_g_recall_precision():
    from EvoDAG import EvoDAG
    y = cl.copy()
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                early_stopping_rounds=100,
                time_limit=0.9,
                multiple_outputs=True,
                seed=0,
                popsize=500)
    gp.y = y
    gp.X = X
    gp.create_population()
    off = gp.random_offspring()
    hy = SparseArray.argmax(off.hy)
    index = np.array(gp._mask_ts.index)
    y = np.array(gp._y_klass.full_array())[index]
    hy = np.array(hy.full_array())[index]
    nclasses = gp._bagging_fitness.nclasses
    precision = np.array([(y[hy == k] == k).mean() for k in range(nclasses)])
    recall = np.array([(hy[y == k] == k).mean() for k in range(nclasses)])
    min_class = gp._bagging_fitness.min_class
    score = (precision[min_class] * recall[min_class]) - 1
    gp._fitness_function = 'g_recall_precision'
    gp._bagging_fitness.set_fitness(off)
    if not np.isfinite(score):
        score = -1
    assert_almost_equals(score, off.fitness)
    index = np.array(gp._mask_ts.full_array()) == 0
    y = np.array(gp._y_klass.full_array())[index]
    hy = SparseArray.argmax(off.hy)
    hy = np.array(hy.full_array())[index]
    precision = np.array([(y[hy == k] == k).mean() for k in range(nclasses)])
    recall = np.array([(hy[y == k] == k).mean() for k in range(nclasses)])
    score = (precision[min_class] * recall[min_class]) - 1
    if not np.isfinite(score):
        score = -1
    assert_almost_equals(score, off.fitness_vs)


def test_g_recall():
    from EvoDAG import EvoDAG
    y = cl.copy()
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                early_stopping_rounds=100,
                time_limit=0.9,
                multiple_outputs=True,
                seed=0,
                popsize=500)
    gp.y = y
    gp.X = X
    gp.create_population()
    off = gp.random_offspring()
    hy = SparseArray.argmax(off.hy)
    index = np.array(gp._mask_ts.index)
    y = np.array(gp._y_klass.full_array())[index]
    hy = np.array(hy.full_array())[index]
    nclasses = gp._bagging_fitness.nclasses
    recall = np.array([(hy[y == k] == k).mean() for k in range(nclasses)])
    score = np.prod(recall) - 1
    gp._fitness_function = 'g_recall'
    gp._bagging_fitness.set_fitness(off)
    assert_almost_equals(score, off.fitness)
    index = np.array(gp._mask_ts.full_array()) == 0
    y = np.array(gp._y_klass.full_array())[index]
    hy = SparseArray.argmax(off.hy)
    hy = np.array(hy.full_array())[index]
    recall = np.array([(hy[y == k] == k).mean() for k in range(nclasses)])
    score = np.prod(recall) - 1
    assert_almost_equals(score, off.fitness_vs)


def test_g_F1():
    from EvoDAG import EvoDAG
    y = cl.copy()
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                early_stopping_rounds=100,
                time_limit=0.9, fitness_function='g_F1',
                multiple_outputs=True, seed=0, popsize=500)
    gp.y = y
    gp.X = X
    gp.create_population()
    off = gp.random_offspring()
    hy = SparseArray.argmax(off.hy)
    index = np.array(gp._mask_ts.index)
    y = np.array(gp._y_klass.full_array())[index]
    hy = np.array(hy.full_array())[index]
    nclasses = gp._bagging_fitness.nclasses
    recall = np.array([(hy[y == k] == k).mean() for k in range(nclasses)])
    precision = np.array([(y[hy == k] == k).mean() for k in range(nclasses)])
    _ = (2 * precision * recall) / (precision + recall)
    m = ~ np.isfinite(_)
    _[m] = 0
    score = np.prod(_) - 1
    assert gp._fitness_function == 'g_F1'
    gp._bagging_fitness.set_fitness(off)
    print(score, _)
    assert_almost_equals(score, off.fitness)
    index = np.array(gp._mask_ts.full_array()) == 0
    y = np.array(gp._y_klass.full_array())[index]
    hy = SparseArray.argmax(off.hy)
    hy = np.array(hy.full_array())[index]
    recall = np.array([(hy[y == k] == k).mean() for k in range(nclasses)])
    precision = np.array([(y[hy == k] == k).mean() for k in range(nclasses)])
    _ = (2 * precision * recall) / (precision + recall)
    m = ~ np.isfinite(_)
    _[m] = 0
    score = np.prod(_) - 1
    print(score, _)
    assert_almost_equals(score, off.fitness_vs)
    # assert False


def test_g_precision():
    from EvoDAG import EvoDAG
    y = cl.copy()
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                early_stopping_rounds=200,
                time_limit=0.9, fitness_function='g_precision',
                multiple_outputs=True, seed=0, popsize=1000)
    gp.y = y
    gp.X = X
    gp.create_population()
    # off = gp.random_offspring()
    off = gp.population.bsf
    hy = SparseArray.argmax(off.hy)
    index = np.array(gp._mask_ts.index)
    y = np.array(gp._y_klass.full_array())[index]
    hy = np.array(hy.full_array())[index]
    nclasses = gp._bagging_fitness.nclasses
    precision = np.array([(y[hy == k] == k).mean() for k in range(nclasses)])
    score = np.prod(precision) - 1
    assert gp._fitness_function == 'g_precision'
    gp._bagging_fitness.set_fitness(off)
    assert_almost_equals(score, off.fitness)
    index = np.array(gp._mask_ts.full_array()) == 0
    y = np.array(gp._y_klass.full_array())[index]
    hy = SparseArray.argmax(off.hy)
    hy = np.array(hy.full_array())[index]
    precision = np.array([(y[hy == k] == k).mean() for k in range(nclasses)])
    score = np.prod(precision) - 1
    if np.isfinite(score) and np.isfinite(off.fitness_vs):
        assert_almost_equals(score, off.fitness_vs)
    # assert False


def test_g_g_recall_precision():
    from EvoDAG import EvoDAG
    y = cl.copy()
    gp = EvoDAG(generations=np.inf,
                tournament_size=2,
                early_stopping_rounds=500,
                time_limit=0.9, fitness_function='g_g_recall_precision',
                multiple_outputs=True, seed=0, popsize=1000)
    gp.y = y
    gp.X = X
    gp.create_population()
    off = gp.random_offspring()
    hy = SparseArray.argmax(off.hy)
    index = np.array(gp._mask_ts.index)
    y = np.array(gp._y_klass.full_array())[index]
    hy = np.array(hy.full_array())[index]
    nclasses = gp._bagging_fitness.nclasses
    precision = np.array([(y[hy == k] == k).mean() for k in range(nclasses)])
    recall = np.array([(hy[y == k] == k).mean() for k in range(nclasses)])
    precision[~ np.isfinite(precision)] = 0
    print(precision, recall)
    score = np.prod([a * b for a, b in zip(recall, precision)]) - 1
    assert gp._fitness_function == 'g_g_recall_precision'
    gp._bagging_fitness.set_fitness(off)
    assert_almost_equals(score, off.fitness)
    index = np.array(gp._mask_ts.full_array()) == 0
    y = np.array(gp._y_klass.full_array())[index]
    hy = SparseArray.argmax(off.hy)
    hy = np.array(hy.full_array())[index]
    precision = np.array([(y[hy == k] == k).mean() for k in range(nclasses)])
    recall = np.array([(hy[y == k] == k).mean() for k in range(nclasses)])
    precision[~ np.isfinite(precision)] = 0
    score = np.prod([a * b for a, b in zip(recall, precision)]) - 1
    assert_almost_equals(score, off.fitness_vs)
    



# Copyright 2015 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
import logging
from SparseArray import SparseArray
from .node import Variable
from .node import Add, Mul, Div, Fabs, Exp, Sqrt, Sin, Cos, Log1p
from .node import Sq, Min, Max
from .node import Atan2, Hypot, Acos, Asin, Atan, Tan, Cosh, Sinh
from .node import Tanh, Acosh, Asinh, Atanh, Expm1, Log, Log2, Log10
from .node import Lgamma, Sign, Ceil, Floor, NaiveBayes, NaiveBayesMN
from .node import Centroid
from .model import Model
from .population import SteadyState
from .utils import tonparray
from .function_selection import FunctionSelection
from .naive_bayes import NaiveBayes as NB
from .bagging_fitness import BaggingFitness
import time
import importlib
import inspect


class EvoDAG(object):
    def __init__(self, generations=np.inf, popsize=10000,
                 seed=0, tournament_size=2,
                 early_stopping_rounds=-1,
                 function_set=[Add, Mul, Div, Fabs,
                               Exp, Sqrt, Sin, Cos, Log1p,
                               Sq, Min, Max, Atan2, Hypot, Acos, Asin, Atan,
                               Tan, Cosh, Sinh, Tanh, Acosh, Asinh, Atanh,
                               Expm1, Log, Log2, Log10, Lgamma, Sign,
                               Ceil, Floor, NaiveBayes, NaiveBayesMN, Centroid],
                 tr_fraction=0.5, population_class=SteadyState,
                 negative_selection=True,
                 number_tries_feasible_ind=30, time_limit=None,
                 unique_individuals=True, classifier=True,
                 labels=None, all_inputs=False, random_generations=0,
                 fitness_function='BER', orthogonal_selection=False, orthogonal_dot_selection=False,
                 min_density=0.8, multiple_outputs=False, function_selection=True,
                 fs_tournament_size=2, finite=True, pr_variable=0.33,
                 share_inputs=False, input_functions=None, F1_index=-1,
                 use_all_vars_input_functions=False, remove_raw_inputs=True,
                 probability_calibration=None, **kwargs):
        self._remove_raw_inputs = remove_raw_inputs
        self._fitness_function = fitness_function
        self._bagging_fitness = BaggingFitness(base=self)
        generations = np.inf if generations is None else generations
        self._pr_variable = pr_variable
        self._share_inputs = share_inputs
        self._finite = finite
        self._generations = generations
        self._popsize = popsize
        self._classifier = classifier
        self._number_tries_feasible_ind = number_tries_feasible_ind
        self._unfeasible_counter = 0
        self._negative_selection = negative_selection
        self._number_tries_unique_args = 3
        self._tr_fraction = tr_fraction
        if early_stopping_rounds is not None and early_stopping_rounds < 0:
            early_stopping_rounds = popsize
        self._early_stopping_rounds = early_stopping_rounds
        self._tournament_size = tournament_size
        self._seed = seed
        self._labels = labels
        self._multiclass = False
        self._function_set = function_set
        self._function_selection = function_selection
        self._fs_tournament_size = fs_tournament_size
        density_safe = [k for k, v in enumerate(function_set) if v.density_safe]
        self._function_selection_ins = FunctionSelection(nfunctions=len(self._function_set),
                                                         seed=seed,
                                                         tournament_size=self._fs_tournament_size,
                                                         nargs=map(lambda x: x.nargs,
                                                                   function_set),
                                                         density_safe=density_safe)
        self._min_density = min_density
        self._function_selection_ins.min_density = self._min_density
        self._time_limit = time_limit
        self._init_time = time.time()
        self._random_generations = random_generations
        if not inspect.isclass(population_class):
            pop = importlib.import_module('EvoDAG.population')
            population_class = getattr(pop, population_class)
        self._set_input_functions(input_functions)
        self._population_class = population_class
        np.random.seed(self._seed)
        self._unique_individuals = unique_individuals
        self._unique_individuals_set = set()
        self._logger = logging.getLogger('EvoDAG')
        self._all_inputs = all_inputs
        if self._generations == np.inf and tr_fraction == 1:
            raise RuntimeError("Infinite evolution, set generations\
            or tr_fraction < 1 ")
        self._multiple_outputs = multiple_outputs
        self._F1_index = F1_index
        self._use_all_vars_input_functions = use_all_vars_input_functions
        self._probability_calibration = probability_calibration
        self._orthogonal_selection = orthogonal_selection
        self._orthogonal_dot_selection = orthogonal_dot_selection
        # orthogonal_selection is only implemented for classification
        # if self._orthogonal_selection:
        #     assert self._classifier
        self._extras = kwargs
        if self._time_limit is not None:
            self._logger.info('Time limit: %0.2f' % self._time_limit)

    def _set_input_functions(self, input_functions):
        if input_functions is not None:
            if not isinstance(input_functions, list):
                input_functions = [input_functions]
            r = []
            for f in input_functions:
                if not inspect.isclass(f):
                    _ = importlib.import_module('EvoDAG.node')
                    f = getattr(_, f)
                    r.append(f)
                else:
                    r.append(f)
            self._input_functions = r
        else:
            self._input_functions = input_functions

    def get_params(self):
        "Parameters used to initialize the class"
        import inspect
        a = inspect.getargspec(self.__init__)[0]
        out = dict()
        for key in a[1:]:
            value = getattr(self, "_%s" % key, None)
            out[key] = value
        return out

    def clone(self):
        "Clone the class without the population"
        return self.__class__(**self.get_params())

    @property
    def classifier(self):
        "Whether it is a classification problem or not"
        return self._classifier

    @property
    def signature(self):
        "Instance file name"
        kw = self.get_params()
        keys = sorted(kw.keys())
        l = []
        for k in keys:
            n = k[0] + k[-1]
            v = kw[k]
            if k == 'function_set':
                v = "_".join([x.__name__[0] +
                              x.__name__[-1] +
                              str(x.nargs) for x in kw[k]])
            elif k == 'population_class':
                v = kw[k].__name__
            else:
                v = str(v)
            l.append('{0}_{1}'.format(n, v))
        return '-'.join(l)

    @property
    def popsize(self):
        """Population size"""
        return self._popsize

    @property
    def generations(self):
        """Number of generations"""
        return self._generations

    @property
    def X(self):
        """Features or variables used in the training and validation set"""
        return self._X

    @X.setter
    def X(self, v):
        self._X = Model.convert_features(v)
        self.nvar = len(self._X)

    @property
    def Xtest(self):
        "Features or variables used in the test set"
        return [x.hy_test for x in self.X]

    @Xtest.setter
    def Xtest(self, v):
        Model.convert_features_test_set(self.X, v)

    @property
    def y(self):
        """Dependent variable"""
        return self._y

    @y.setter
    def y(self, v):
        if isinstance(v, np.ndarray):
            v = SparseArray.fromlist(v)
        if self.classifier:
            self._multiple_outputs = True
            if self._labels is None:
                self.nclasses(v)
            return self._bagging_fitness.multiple_outputs_cl(v)
        elif self._multiple_outputs:
            return self._bagging_fitness.multiple_outputs_regression(v)
        elif self._tr_fraction < 1:
            for i in range(self._number_tries_feasible_ind):
                self._bagging_fitness.set_regression_mask(v)
                flag = self._bagging_fitness.test_regression_mask(v)
                if flag:
                    break
            if not flag:
                msg = "Unsuitable validation set (RSE: average equals zero)"
                raise RuntimeError(msg)
        else:
            self._mask = 1.0
        self._ytr = v * self._mask
        self._y = v

    @property
    def function_set(self):
        "List containing the functions used to create the individuals"
        return self._function_set

    @property
    def nvar(self):
        """Number of features or variables"""
        return self._nvar

    @nvar.setter
    def nvar(self, v):
        self._nvar = v

    @property
    def naive_bayes(self):
        try:
            return self._naive_bayes
        except AttributeError:
            if hasattr(self, '_y_klass'):
                self._naive_bayes = NB(mask=self._mask_ts, klass=self._y_klass,
                                       nclass=self._labels.shape[0])
            else:
                self._naive_bayes = None
        return self._naive_bayes

    @property
    def population(self):
        "Class containing the population and all the individuals generated"
        try:
            return self._p
        except AttributeError:
            self._p = self._population_class(base=self,
                                             tournament_size=self._tournament_size,
                                             classifier=self.classifier,
                                             labels=self._labels,
                                             es_extra_test=self.es_extra_test,
                                             popsize=self._popsize,
                                             random_generations=self._random_generations,
                                             negative_selection=self._negative_selection)
            return self._p

    @property
    def init_popsize(self):
        if self._all_inputs:
            return self._init_popsize
        return self.popsize

    def es_extra_test(self, v):
        """This function is called from population before setting
        the early stopping individual and after the comparisons with
        the validation set fitness"""
        return True

    def _random_leaf(self, var):
        v = Variable(var, ytr=self._ytr, finite=self._finite, mask=self._mask,
                     naive_bayes=self.naive_bayes)
        if not v.eval(self.X):
            return None
        if not v.isfinite():
            return None
        if not self._bagging_fitness.set_fitness(v):
            return None
        return v

    def random_leaf(self):
        "Returns a random variable with the associated weight"
        for i in range(self._number_tries_feasible_ind):
            var = np.random.randint(self.nvar)
            v = self._random_leaf(var)
            if v is None:
                continue
            return v
        raise RuntimeError("Could not find a suitable random leaf")

    def unique_individual(self, v):
        "Test whether v has not been explored during the evolution"
        return v not in self._unique_individuals_set

    def unfeasible_offspring(self):
        self._unfeasible_counter += 1
        return None

    def _random_offspring(self, func, args):
        f = func(args, ytr=self._ytr, naive_bayes=self.naive_bayes,
                 finite=self._finite, mask=self._mask)
        if self._unique_individuals:
            sig = f.signature()
            if self.unique_individual(sig):
                self._unique_individuals_set.add(sig)
            else:
                return self.unfeasible_offspring()
        f.height = max([self.population.hist[x].height for x in args]) + 1
        if not f.eval(self.population.hist):
            return self.unfeasible_offspring()
        if not f.isfinite():
            return self.unfeasible_offspring()
        if not self._bagging_fitness.set_fitness(f):
            return self.unfeasible_offspring()
        return f

    def _get_args_orthogonal(self, first):
        vars = self.population.random()
        pop = self.population.population
        score = self._bagging_fitness.score
        if self.classifier:
            fit = [(k, score.accuracy(first, SparseArray.argmax(pop[x].hy),
                                      self._mask_ts.index)[0]) for k, x in enumerate(vars)]
        else:
            fit = [(k, score.accuracy(first, pop[x].hy.sign() + 1,
                                      self._mask.index)[0]) for k, x in enumerate(vars)]
        fit = min(fit, key=lambda x: x[1])
        index = fit[0]
        return vars[index]

    def get_args_orthogonal(self, func):
        first = self.population.tournament()
        args = {first: 1}
        if self.classifier:
            first = SparseArray.argmax(self.population.population[first].hy)
        else:
            first = self.population.population[first].hy.sign() + 1.0
        res = []
        sel = self._get_args_orthogonal
        n_tries = self._number_tries_unique_args
        for j in range(func.nargs - 1):
            k = sel(first)
            for _ in range(n_tries):
                if k not in args:
                    args[k] = 1
                    res.append(k)
                    break
                else:
                    k = sel(first)
        if len(res) < func.min_nargs:
            return None
        return res

    def _get_args_orthogonal_dot(self, first_hy):    
        vars = self.population.random()
        pop = self.population.population
        if self.classifier:
            mask = self._mask_ts
        else:
            mask = self._mask
            
        if isinstance(first_hy,list):
            prod = []
            for k,x in enumerate(vars):  
                pvalue = 0
                for i in range(len(first_hy)):
                    pvalue+= SparseArray.dot(pop[x].hy[i].mul(mask),first_hy[i])
                prod.append( (k,pvalue) )
        else:
            prod = [(k,SparseArray.dot(pop[x].hy.mul(mask),first_hy)) for k,x in enumerate(vars)]
        
        prod = min(prod, key=lambda x: x[1])
        index = prod[0]
        return vars[index]
    
    def get_args_orthogonal_dot(self, func):
        first = self.population.tournament()
        args = {first: 1}
        if self.classifier:
            mask = self._mask_ts
        else:
            mask = self._mask
        if isinstance(self.population.population[first].hy,list):
            first_hy = []
            for i in range(len(self.population.population[first].hy)):
                first_hy.append( self.population.population[first].hy[i].mul(mask) )
        else:
            first_hy = self.population.population[first].hy.mul(mask)
            
        res = []
        sel = self._get_args_orthogonal_dot
        n_tries = self._number_tries_unique_args
        for j in range(func.nargs - 1):
            k = sel(first_hy)
            for _ in range(n_tries):
                if k not in args:
                    args[k] = 1
                    res.append(k)
                    break
                else:
                    k = sel(first_hy)
        try:
            min_nargs = func.min_nargs
        except AttributeError:
            min_nargs = func.nargs
        if len(res) < min_nargs:
            return None
        return res
    
    def get_unique_args(self, func):
        args = {}
        res = []
        p_tournament = self.population.tournament
        n_tries = self._number_tries_unique_args
        for j in range(func.nargs):
            k = p_tournament()
            for _ in range(n_tries):
                if k not in args:
                    args[k] = 1
                    res.append(k)
                    break
                else:
                    k = p_tournament()
        if len(res) < func.min_nargs:
            return None
        return res

    def get_args(self, func):
        args = []
        if self._orthogonal_selection and func.orthogonal_selection:
            return self.get_args_orthogonal(func)
        if self._orthogonal_dot_selection and func.orthogonal_selection:
            return self.get_args_orthogonal_dot(func)
        if func.unique_args:
            return self.get_unique_args(func)
        try:
            min_nargs = func.min_nargs
        except AttributeError:
            min_nargs = func.nargs
        p_tournament = self.population.tournament
        for j in range(func.nargs):
            k = p_tournament()
            for _ in range(self._number_tries_unique_args):
                if k not in args:
                    break
                else:
                    k = p_tournament()
            args.append(k)
        if len(args) < min_nargs:
            return None
        return args

    def random_offspring(self):
        "Returns an offspring with the associated weight(s)"
        function_set = self.function_set
        function_selection = self._function_selection_ins
        function_selection.density = self.population.density
        function_selection.unfeasible_functions.clear()
        for i in range(self._number_tries_feasible_ind):
            if self._function_selection:
                func_index = function_selection.tournament()
            else:
                func_index = function_selection.random_function()
            func = function_set[func_index]
            args = self.get_args(func)
            if args is None:
                continue
            args = [self.population.population[x].position for x in args]
            f = self._random_offspring(func, args)
            if f is None:
                function_selection.unfeasible_functions.add(func_index)
                continue
            function_selection[func_index] = f.fitness
            return f
        raise RuntimeError("Could not find a suitable random offpsring")

    def create_population(self):
        "Create the initial population"
        self.population.create_population()

    def stopping_criteria_tl(self):
        if self._time_limit is not None:
            if time.time() - self._init_time > self._time_limit:
                return True
        return False

    def stopping_criteria(self):
        "Test whether the stopping criteria has been achieved."
        if self.stopping_criteria_tl():
            return True
        if self.generations < np.inf:
            inds = self.popsize * self.generations
            flag = inds <= len(self.population.hist)
        else:
            flag = False
        if flag:
            return True
        est = self.population.estopping
        if self._tr_fraction < 1:
            if est is not None and est.fitness_vs == 0:
                return True
        esr = self._early_stopping_rounds
        if self._tr_fraction < 1 and esr is not None and est is not None:
            position = self.population.estopping.position
            if position < self.init_popsize:
                position = self.init_popsize
            return (len(self.population.hist) +
                    self._unfeasible_counter -
                    position) > esr
        return flag

    def nclasses(self, v):
        "Number of classes of v, also sets the labes"
        if not self.classifier:
            return 0
        if isinstance(v, list):
            self._labels = np.arange(len(v))
            return
        if not isinstance(v, np.ndarray):
            v = tonparray(v)
        self._labels = np.unique(v)
        return self._labels.shape[0]

    def replace(self, a):
        "Replace an individual in the population with individual a"
        self.population.replace(a)

    def shuffle_tr2ts(self):
        L = []
        for x in self.X:
            l = np.array(x.hy.full_array())
            np.random.shuffle(l)
            L.append(l)
        return np.array(L).T

    def fit(self, X, y, test_set=None):
        """Evolutive process"""
        self._init_time = time.time()
        self.X = X
        if self._popsize == "nvar":
            self._popsize = self.nvar + len(self._input_functions)
        if isinstance(test_set, str) and test_set == 'shuffle':
            test_set = self.shuffle_tr2ts()
        nclasses = self.nclasses(y)
        if self.classifier and self._multiple_outputs:
            pass
        elif nclasses > 2:
            assert False
            self._multiclass = True
            return self.multiclass(X, y, test_set=test_set)
        self.y = y
        if test_set is not None:
            self.Xtest = test_set
        for _ in range(self._number_tries_feasible_ind):
            self._logger.info("Starting evolution")
            try:
                self.create_population()
                if self.stopping_criteria_tl():
                    break
            except RuntimeError as err:
                self._logger.info("Done evolution (RuntimeError (%s), hist: %s)" % (err, len(self.population.hist)))
                return self
            self._logger.info("Population created (hist: %s)" % len(self.population.hist))
            if len(self.population.hist) >= self._tournament_size:
                break
        if len(self.population.hist) == 0:
            raise RuntimeError("Could not find a suitable individual")
        if len(self.population.hist) < self._tournament_size:
            self._logger.info("Done evolution (hist: %s)" % len(self.population.hist))
            return self
        if self._remove_raw_inputs:
            for x in range(self.nvar):
                self._X[x] = None
        while not self.stopping_criteria():
            try:
                a = self.random_offspring()
            except RuntimeError as err:
                self._logger.info("Done evolution (RuntimeError (%s), hist: %s)" % (err, len(self.population.hist)))
                return self
            self.replace(a)
        self._logger.info("Done evolution (hist: %s)" % len(self.population.hist))
        return self

    def trace(self, n):
        "Restore the position in the history of individual v's nodes"
        return self.population.trace(n)

    def model(self, v=None):
        "Returns the model of node v"
        return self.population.model(v=v)

    def decision_function(self, v=None, X=None):
        "Decision function i.e. the raw data of the prediction"
        m = self.model(v=v)
        return m.decision_function(X)

    def predict(self, v=None, X=None):
        """In classification this returns the classes, in
        regression it is equivalent to the decision function"""
        if X is None:
            X = v
            v = None
        m = self.model(v=v)
        return m.predict(X)

    @classmethod
    def init(cls, params_fname=None, seed=None, classifier=True, **kwargs):
        import os
        from .utils import RandomParameterSearch
        import json
        import gzip
        if params_fname is None:
            if classifier:
                kw = os.path.join(os.path.dirname(__file__),
                                  'conf', 'default_parameters.json')
            else:
                kw = os.path.join(os.path.dirname(__file__),
                                  'conf', 'default_parameters_r.json')
        else:
            kw = params_fname
        if not isinstance(kw, dict):
            if kw.endswith('.gz'):
                func = gzip.open
            else:
                func = open
            with func(kw, 'rb') as fpt:
                try:
                    d = fpt.read()
                    kw = json.loads(str(d, encoding='utf-8'))
                except TypeError:
                    kw = json.loads(d)
            if isinstance(kw, list):
                kw = kw[0]
        if seed is not None:
            kw['seed'] = seed
        kw.update(kwargs)
        kw = RandomParameterSearch.process_params(kw)
        return cls(**kw)


RGP = EvoDAG
RootGP = RGP



# Copyright 2015 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
from SparseArray import SparseArray
from .node import Variable, Function
from .utils import tonparray
from multiprocessing import Pool
import logging
import gc
import os
import gzip
import pickle
import shutil
from time import time
try:
    from tqdm import tqdm
except ImportError:
    def tqdm(x, **kwargs):
        return x
LOGGER = logging.getLogger('EvoDAG')


def fit(X_y_evodag):
    X, y, test_set, evodag, tmpdir, init_time = X_y_evodag
    if tmpdir is not None:
        seed = evodag['seed']
        output = os.path.join(tmpdir, '%s.evodag' % seed)
        if os.path.isfile(output):
            with gzip.open(output) as fpt:
                try:
                    return pickle.load(fpt)
                except Exception:
                    pass
    try:
        time_limit = evodag['time_limit']
        if time_limit is not None:
            evodag['time_limit'] = time_limit - (time() - init_time)
            if evodag['time_limit'] < 2:
                LOGGER.info('Not enough time (seed: %s) ' % evodag['seed'])
                return None
    except KeyError:
        pass
    try:
        evodag = EvoDAG(**evodag)
        evodag.fit(X, y, test_set=test_set)
    except RuntimeError:
        return None
    m = evodag.model
    gc.collect()
    if tmpdir is not None:
        with gzip.open(output, 'w') as fpt:
            pickle.dump(m, fpt)
    return m


def decision_function(model_X):
    k, model, X = model_X
    return [k, model.decision_function(X)]


def predict_proba(model_X):
    k, model, X = model_X
    return [k, model.predict_proba(X)]


class Model(object):
    """Object to store the necesary elements to make predictions
    based on an individual"""
    def __init__(self, trace, hist, nvar=None, classifier=True, labels=None,
                 probability_calibration=None, nclasses=None):
        self._classifier = classifier
        self._trace = trace
        self._hist = hist
        self._map = {}
        for k, v in enumerate(self._trace):
            self._map[v] = k
        self._hy_test = self._hist[self._trace[-1]].hy_test
        self._hist = [self.transform(self._hist[x].tostore()) for x in
                      self._trace]
        self._labels = labels
        self._nvar = nvar
        self._probability_calibration = probability_calibration
        self._nclasses = nclasses
        
    @property
    def nclasses(self):
        return self._nclasses

    @property
    def nvar(self):
        return self._nvar

    @property
    def multiple_outputs(self):
        return self._hist[0]._multiple_outputs

    @property
    def classifier(self):
        "whether this is classification or regression task"
        return self._classifier

    @property
    def fitness_vs(self):
        "Fitness in the validation set"
        return self._hist[-1].fitness_vs

    @property
    def size(self):
        return len(self._hist)

    @property
    def height(self):
        return self._hist[-1].height

    def inputs(self, counter=None):
        from collections import Counter
        if counter is None:
            counter = Counter()
        for node in self._hist:
            if node.height == 0:
                if isinstance(node._variable, list):
                    for _ in node._variable:
                        counter[_] += 1
                else:
                    counter[node._variable] += 1
        return counter

    def transform(self, v):
        if v.height == 0:
            return v
        if v.nargs == 1:
            v.variable = self._map[v.variable]
        else:
            v.variable = [self._map[x] for x in v.variable]
        return v

    def predict_proba(self, X, **kwargs):
        X = self.decision_function(X, **kwargs)
        return self._probability_calibration.predict_proba(X)

    def decision_function(self, X, **kwargs):
        "Decision function i.e. the raw data of the prediction"
        if X is None:
            return self._hy_test
        X = self.convert_features(X)
        if len(X) < self.nvar:
            _ = 'Number of variables differ, trained with %s given %s' % (self.nvar, len(X))
            raise RuntimeError(_)
        hist = self._hist
        for node in hist:
            if node.height:
                node.eval(hist)
            else:
                node.eval(X)
        node.normalize()
        r = node.hy
        for i in hist[:-1]:
            i.hy = None
            i.hy_test = None
        gc.collect()
        return r

    def predict(self, X, **kwargs):
        hy = self.decision_function(X, **kwargs)
        if self._classifier:
            [x.finite(inplace=True) for x in hy]
            hy = np.array(SparseArray.argmax(hy).full_array(), dtype=np.int)
            if self._labels is not None:
                hy = self._labels[hy]
        else:
            hy = tonparray(hy)
        return hy

    def graphviz(self, fpt, terminals=True):
        flag = False
        if isinstance(fpt, str):
            flag = True
            fpt = open(fpt, 'w')
        fpt.write("digraph EvoDAG {\n")
        last = len(self._hist) - 1
        height = self._hist[-1].height
        try:
            b, m = np.linalg.solve([[0, height-1], [1, 1]], [9, 1])
        except np.linalg.linalg.LinAlgError:
            b, m = 0, 1
        done = {}
        for k, n in enumerate(self._hist):
            if isinstance(n, Function):
                done[k] = 1
                name = n.__class__.__name__
                if n.height == 0:
                    cdn = "n{0} [label=\"{1}\" fillcolor=red style=filled];\n"
                    fpt.write(cdn.format(k, name))
                    continue
                color = int(np.round(n.height * m + b))
                extra = "colorscheme=blues9 style=filled color={0}".format(color)
                if k == last:
                    extra = "fillcolor=green style=filled"
                fpt.write("n{0} [label=\"{1}\" {2}];\n".format(k,
                                                               name,
                                                               extra))
                vars = n._variable
                if not isinstance(vars, list):
                    vars = [vars]
                for j in vars:
                    if j in done:
                        fpt.write("n{0} -> n{1};\n".format(k, j))
            elif terminals:
                cdn = "n{0} [label=\"X{1}\" fillcolor=red style=filled];\n"
                fpt.write(cdn.format(k, n._variable))
                done[k] = 1
        fpt.write("}\n")
        if flag:
            fpt.close()

    @staticmethod
    def convert_features(v):
        if v is None:
            return None
        if isinstance(v[0], Variable):
            return v
        if isinstance(v, np.ndarray):
            X = v.T
        elif isinstance(v[0], SparseArray):
            X = v
        else:
            X = np.array(v).T
        lst = []
        for var, d in enumerate(X):
            v = Variable(var, 1)
            if isinstance(d, SparseArray):
                v._eval_tr = d
            else:
                v._eval_tr = SparseArray.fromlist(d)
            lst.append(v)
        return lst

    @staticmethod
    def convert_features_test_set(vars, v):
        if isinstance(v, np.ndarray):
            X = v.T
        else:
            X = v
        for var, d in zip(vars, X):
            if isinstance(d, SparseArray):
                var._eval_ts = d
            else:
                var._eval_ts = SparseArray.fromlist(d)


class Ensemble(object):
    "Ensemble that predicts using the average"
    def __init__(self, models, n_jobs=1, evodags=None, tmpdir=None):
        self._models = models
        self._n_jobs = n_jobs
        self._evodags = evodags
        self._tmpdir = tmpdir
        if models is not None:
            self._init()

    def fit(self, X, y, test_set=None):
        evodags = self._evodags
        init_time = time()
        args = [(X, y, test_set, evodag, self._tmpdir, init_time) for evodag in evodags]
        try:
            time_limit = evodags[0]['time_limit']
        except KeyError:
            time_limit = None
        if time_limit is not None:
            LOGGER.info('time_limit in Ensemble: %0.2f' % time_limit)
        if self._n_jobs == 1:
            _ = [fit(x) for x in tqdm(args)]
            self._models = [x for x in _ if x is not None]
        else:
            p = Pool(self._n_jobs, maxtasksperchild=1)
            self._models = []
            for x in tqdm(p.imap_unordered(fit, args),
                          total=len(args)):
                if x is not None:
                    self._models.append(x)
                if time_limit is not None and time() - init_time > time_limit:
                    p.terminate()
                    break
            p.close()
        if self._tmpdir is not None:
            shutil.rmtree(self._tmpdir)
        self._init()
        if time_limit is not None:
            LOGGER.info('Used time in Ensemble: %0.2f' % (time() - init_time))
        return self

    def _init(self):
        self._labels = self._models[0]._labels
        self._classifier = False
        flag = False
        if self._models[0]._classifier:
            flag = True
        self._classifier = flag

    @property
    def nclasses(self):
        return self.models[0].nclasses

    @property
    def probability_calibration(self):
        return self.models[0]._probability_calibration is not None

    @property
    def models(self):
        "List containing the models that compose the ensemble"
        return self._models

    @property
    def multiple_outputs(self):
        return self.models[0].multiple_outputs

    @property
    def classifier(self):
        return self._classifier

    @property
    def fitness_vs(self):
        "Median Fitness in the validation set"
        l = [x.fitness_vs for x in self.models]
        return np.median(l)

    @property
    def size(self):
        l = [x.size for x in self.models]
        return np.median(l)

    @property
    def height(self):
        l = [x.height for x in self.models]
        return np.median(l)

    def inputs(self, counter=None):
        from collections import Counter
        if counter is None:
            counter = Counter()
        for m in self.models:
            m.inputs(counter=counter)
        return counter

    def _decision_function_raw(self, X, cpu_cores=1):
        if cpu_cores == 1:
            r = [m.decision_function(X) for m in self._models]
        else:
            p = Pool(cpu_cores, maxtasksperchild=1)
            args = [(k, m, X) for k, m in enumerate(self._models)]
            r = [x for x in tqdm(p.imap_unordered(decision_function,
                                                  args),
                                 total=len(args))]
            r.sort(key=lambda x: x[0])
            r = [x[1] for x in r]
            p.close()
        return r

    def raw_decision_function(self, X):
        hy = self._decision_function_raw(X, cpu_cores=self._n_jobs)
        if isinstance(hy[0], list):
            _ = []
            [[_.append(y) for y in x] for x in hy]
            hy = _
        if self.classifier:
            [x.finite(inplace=True) for x in hy]
        return np.array([tonparray(x) for x in hy]).T

    def _predict_proba_raw(self, X, cpu_cores=1):
        if cpu_cores == 1:
            r = [m.predict_proba(X) for m in self._models]
        else:
            p = Pool(cpu_cores, maxtasksperchild=1)
            args = [(k, m, X) for k, m in enumerate(self._models)]
            r = [x for x in tqdm(p.imap_unordered(predict_proba,
                                                  args),
                                 total=len(args))]
            r.sort(key=lambda x: x[0])
            r = [x[1] for x in r]
            p.close()
        return r

    def predict_proba(self, X):
        if self.probability_calibration:
            proba = np.array(self._predict_proba_raw(X, cpu_cores=self._n_jobs))
            proba = np.mean(proba, axis=0)
            proba /= np.sum(proba, axis=1)[:, np.newaxis]
            proba[np.isnan(proba)] = 1. / self.nclasses
            return proba
        hy = self._decision_function_raw(X, cpu_cores=self._n_jobs)
        minlength = len(hy[0])
        hy = [SparseArray.argmax(x) for x in hy]
        hy = np.array([x.full_array() for x in hy], dtype=np.int).T
        hy = [np.bincount(x, minlength=minlength) for x in hy]
        return np.array([x / float(x.sum()) for x in hy])

    def decision_function(self, X, cpu_cores=1):
        cpu_cores = max(cpu_cores, self._n_jobs)
        r = self._decision_function_raw(X, cpu_cores=cpu_cores)
        if isinstance(r[0], SparseArray):
            r = np.array([tonparray(x) for x in r if x.isfinite()])
            r = np.median(r, axis=0)
        else:
            [[x.finite(inplace=True) for x in o] for o in r]
            r = np.array([[tonparray(y) for y in x] for x in r])
            r = np.median(r, axis=0)
        return r.T

    def predict(self, X, cpu_cores=1):
        cpu_cores = max(cpu_cores, self._n_jobs)
        if self.classifier:
            return self.predict_cl(X, cpu_cores=cpu_cores)
        return self.decision_function(X, cpu_cores=cpu_cores)

    def predict_cl(self, X, cpu_cores=1):
        cpu_cores = max(cpu_cores, self._n_jobs)
        hy = [SparseArray.argmax(x) for x in
              self._decision_function_raw(X, cpu_cores=cpu_cores)]
        hy = np.array([x.full_array() for x in hy], dtype=np.int).T
        hy = [np.bincount(x).argmax() for x in hy]
        if self._labels is not None:
            hy = self._labels[hy]
        return hy

    def graphviz(self, directory, **kwargs):
        "Directory to store the graphviz models"
        import os
        if not os.path.isdir(directory):
            os.mkdir(directory)
        output = os.path.join(directory, 'evodag-%s')
        for k, m in enumerate(self.models):
            m.graphviz(output % k, **kwargs)

    @classmethod
    def init(cls, n_estimators=30, n_jobs=1, tmpdir=None, **kwargs):
        try:
            init_seed = kwargs['seed']
            del kwargs['seed']
        except KeyError:
            init_seed = 0
        lst = []
        for x in range(init_seed, init_seed + n_estimators):
            kwargs['seed'] = x
            lst.append(kwargs.copy())
        if tmpdir is not None and not os.path.isdir(tmpdir):
            os.mkdir(tmpdir)
        return cls(None, evodags=lst, n_jobs=n_jobs, tmpdir=tmpdir)


class EvoDAGE(object):
    def __init__(self, time_limit=None, **kwargs):
        self._m = Ensemble.init(time_limit=time_limit, **kwargs)
        self._time_limit = time_limit

    @property
    def time_limit(self):
        return self._time_limit

    @time_limit.setter
    def time_limit(self, time_limit):
        self._time_limit = time_limit
        for x in self._m._evodags:
            x['time_limit'] = self._time_limit

    def fit(self, *args, **kwargs):
        return self._m.fit(*args, **kwargs)

    def predict(self, *args, **kwargs):
        return self._m.predict(*args, **kwargs)

    def decision_function(self, *args, **kwargs):
        return self._m.decision_function(*args, **kwargs)

    def raw_decision_function(self, *args, **kwargs):
        return self._m.raw_decision_function(*args, **kwargs)

    def predict_proba(self, *args, **kwargs):
        return self._m.predict_proba(*args, **kwargs)


class EvoDAG(EvoDAGE):
    def __init__(self, **kwargs):
        from EvoDAG import EvoDAG as evodag
        self._m = evodag.init(**kwargs)

    def fit(self, *args, **kwargs):
        self._m.fit(*args, **kwargs)
        self._m = self._m.model()
        return self

    @property
    def model(self):
        return self._m



# Copyright 2015 Mario Graff Guerrero

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import numpy as np
from SparseArray import SparseArray
from .utils import tonparray
from .cython_utils import fitness_SAE, Score


class BaggingFitness(object):
    def __init__(self, base=None):
        self._base = base
        self.assert_fitness_function()

    def assert_fitness_function(self):
        base = self._base
        assert base._fitness_function in ['ER', 'macro-F1', 'BER', 'macro-Precision',
                                          'DotF1', 'DotRecall', 'DotPrecision',
                                          'DotRecallDotPrecision', 'F1',
                                          'RecallDotPrecision', 'macro-RecallF1',
                                          'accDotMacroF1',
                                          'accuracy', 'g_recall_precision',
                                          'g_precision', 'a_precision', 'g_recall',
                                          'g_g_recall_precision', 'g_F1', 'a_F1']

    @property
    def nclasses(self):
        try:
            return self._nclasses
        except AttributeError:
            if self._base.classifier:
                self._nclasses = self._base._labels.shape[0]
            elif self._base._orthogonal_selection:
                self._nclasses = 3
        return self._nclasses

    @property
    def score(self):
        try:
            return self._score
        except AttributeError:
            self._score = Score(self.nclasses)
        return self._score

    @property
    def min_class(self):
        try:
            return self._min_class
        except AttributeError:
            klass = tonparray(self._base._y_klass)
            c = [(klass == x).sum() for x in np.unique(klass)]
            self._min_class = np.argmin(c)
        return self._min_class

    def mask_fitness_BER(self, k):
        base = self._base
        k = k.argmax(axis=1)
        base._y_klass = SparseArray.fromlist(k)
        klass = np.unique(k)
        cnt = np.min([(k == x).sum() for x in klass]) * (1 - base._tr_fraction)
        cnt = int(np.floor(cnt))
        if cnt == 0:
            cnt = 1
        mask = np.ones_like(k, dtype=np.bool)
        mask_ts = np.zeros(k.shape[0])
        for i in klass:
            index = np.where(k == i)[0]
            np.random.shuffle(index)
            mask[index[:cnt]] = False
            mask_ts[index[cnt:]] = 1.0 / (1.0 * index[cnt:].shape[0] * klass.shape[0])
        base._mask_vs = SparseArray.fromlist(~mask)
        base._mask_ts = SparseArray.fromlist(mask_ts)
        return mask

    def mask_fitness_function(self, k):
        base = self._base
        if base._fitness_function == 'BER':
            return self.mask_fitness_BER(k)
        elif base._fitness_function == 'macro-RecallF1':
            return self.mask_fitness_BER(k)
        k = k.argmax(axis=1)
        base._y_klass = SparseArray.fromlist(k)
        cnt = k.shape[0] * (1 - base._tr_fraction)
        cnt = int(np.floor(cnt))
        if cnt == 0:
            cnt = 1
        mask = np.ones_like(k, dtype=np.bool)
        mask_ts = np.zeros(k.shape[0])
        index = np.arange(k.shape[0])
        nklass = np.unique(k).shape[0]
        for i in range(base._number_tries_feasible_ind):
            np.random.shuffle(index)
            _count = np.unique(k[index[cnt:]], return_counts=True)[1]
            if _count.shape[0] == nklass and np.all(_count > 0):
                break
        mask[index[:cnt]] = False
        mask_ts[index[cnt:]] = 1.0
        base._mask_vs = SparseArray.fromlist(~mask)
        base._mask_ts = SparseArray.fromlist(mask_ts / mask_ts.sum())
        return mask

    def set_classifier_mask(self, v, base_mask=True):
        """Computes the mask used to create the training and validation set"""
        base = self._base
        v = tonparray(v)
        a = np.unique(v)
        if a[0] != -1 or a[1] != 1:
            raise RuntimeError("The labels must be -1 and 1 (%s)" % a)
        mask = np.zeros_like(v)
        cnt = min([(v == x).sum() for x in a]) * base._tr_fraction
        cnt = int(round(cnt))
        for i in a:
            index = np.where((v == i) & base_mask)[0]
            np.random.shuffle(index)
            mask[index[:cnt]] = True
        base._mask = SparseArray.fromlist(mask)
        return SparseArray.fromlist(v)

    def transform_to_mo(self, v):
        base = self._base
        klass = base._labels
        y = np.empty((v.shape[0], klass.shape[0]))
        y.fill(-1)
        for i, k in enumerate(klass):
            mask = k == v
            y[mask, i] = 1
        return y

    def multiple_outputs_cl(self, v):
        base = self._base
        if isinstance(v, list):
            assert len(v) == base._labels.shape[0]
            v = np.array([tonparray(x) for x in v]).T
        else:
            v = tonparray(v)
            v = self.transform_to_mo(v)
        base_mask = self.mask_fitness_function(v)
        mask = []
        ytr = []
        y = []
        for _v in v.T:
            _v = SparseArray.fromlist(_v)
            self.set_classifier_mask(_v, base_mask)
            mask.append(base._mask)
            ytr.append(_v * base._mask)
            y.append(_v)
            base._y = _v
        base._ytr = ytr
        base._y = y
        base._mask = mask

    def set_regression_mask(self, v):
        """Computes the mask used to create the training and validation set"""
        base = self._base
        index = np.arange(v.size())
        np.random.shuffle(index)
        ones = np.ones(v.size())
        ones[index[int(base._tr_fraction * v.size()):]] = 0
        base._mask = SparseArray.fromlist(ones)

    def test_regression_mask(self, v):
        """Test whether the average prediction is different than zero"""
        base = self._base
        m = (base._mask + -1.0).fabs()
        x = v * m
        b = (x + -x.sum() / x.size()).sq().sum()
        return b != 0

    def multiple_outputs_regression(self, v):
        base = self._base
        assert isinstance(v, list)
        v = np.array([tonparray(x) for x in v]).T
        mask = []
        ytr = []
        y = []
        for _v in v.T:
            _v = SparseArray.fromlist(_v)
            for _ in range(base._number_tries_feasible_ind):
                self.set_regression_mask(_v)
                flag = self.test_regression_mask(_v)
                if flag:
                    break
            if not flag:
                msg = "Unsuitable validation set (RSE: average equals zero)"
                raise RuntimeError(msg)
            mask.append(base._mask)
            ytr.append(_v * base._mask)
            y.append(_v)
            base._y = _v
        base._ytr = ytr
        base._y = y
        base._mask = mask

    def del_error(self, v):
        try:
            delattr(v, '_error')
        except AttributeError:
            pass

    def fitness(self, v):
        "Fitness function in the training set"
        base = self._base
        if base._classifier:
            if base._multiple_outputs:
                hy = SparseArray.argmax(v.hy)
                fit_func = base._fitness_function
                if fit_func == 'macro-F1' or fit_func == 'a_F1':
                    f1_score = self.score
                    mf1, mf1_v = f1_score.a_F1(base._y_klass, hy, base._mask_ts.index)
                    v._error = mf1_v - 1
                    v.fitness = mf1 - 1
                elif fit_func == 'DotF1' or fit_func == 'g_F1':
                    f1_score = self.score
                    mf1, mf1_v = f1_score.g_F1(base._y_klass, hy, base._mask_ts.index)
                    v._error = mf1_v - 1
                    v.fitness = mf1 - 1
                elif fit_func == 'DotRecallDotPrecision' or fit_func == 'g_g_recall_precision':
                    f1_score = self.score
                    mf1, mf1_v = f1_score.g_g_recall_precision(base._y_klass, hy,
                                                               base._mask_ts.index)
                    v._error = mf1_v - 1
                    v.fitness = mf1 - 1
                elif fit_func == 'BER' or fit_func == 'a_recall':
                    f1_score = self.score
                    mf1, mf1_v = f1_score.a_recall(base._y_klass, hy, base._mask_ts.index)
                    v._error = mf1_v - 1
                    v.fitness = mf1 - 1
                elif fit_func == 'DotRecall' or fit_func == 'g_recall':
                    f1_score = self.score
                    mf1, mf1_v = f1_score.g_recall(base._y_klass, hy,
                                                   base._mask_ts.index)
                    v._error = mf1_v - 1
                    v.fitness = mf1 - 1
                elif fit_func == 'macro-Precision' or fit_func == 'a_precision':
                    f1_score = self.score
                    mf1, mf1_v = f1_score.a_precision(base._y_klass, hy,
                                                      base._mask_ts.index)
                    v._error = mf1_v - 1
                    v.fitness = mf1 - 1
                elif fit_func == 'DotPrecision' or fit_func == 'g_precision':
                    f1_score = self.score
                    mf1, mf1_v = f1_score.g_precision(base._y_klass, hy,
                                                      base._mask_ts.index)
                    v._error = mf1_v - 1
                    v.fitness = mf1 - 1
                elif fit_func == 'accDotMacroF1':
                    f1_score = self.score
                    mf1, mf1_v = f1_score.accDotMacroF1(base._y_klass, hy,
                                                        base._mask_ts.index)
                    v._error = mf1_v - 1
                    v.fitness = mf1 - 1
                elif fit_func == 'macro-RecallF1':
                    f1_score = self.score
                    mf1, mf1_v = f1_score.macroRecallF1(base._y_klass, hy,
                                                        base._mask_ts.index)
                    v._error = mf1_v - 1
                    v.fitness = mf1 - 1
                elif fit_func == 'F1':
                    f1_score = self.score
                    f1_index = self._base._F1_index
                    index = self.min_class if f1_index < 0 else f1_index
                    mf1, mf1_v = f1_score.F1(index, base._y_klass,
                                             hy, base._mask_ts.index)
                    v._error = mf1_v - 1
                    v.fitness = mf1 - 1
                elif fit_func == 'RecallDotPrecision' or fit_func == 'g_recall_precision':
                    f1_score = self.score
                    mf1, mf1_v = f1_score.g_recall_precision(self.min_class,
                                                             base._y_klass,
                                                             hy, base._mask_ts.index)
                    v._error = mf1_v - 1
                    v.fitness = mf1 - 1
                elif fit_func == 'ER' or fit_func == 'accuracy':
                    f1_score = self.score
                    mf1, mf1_v = f1_score.accuracy(base._y_klass,
                                                   hy, base._mask_ts.index)
                    v._error = mf1_v - 1
                    v.fitness = mf1 - 1
                else:
                    raise RuntimeError('Unknown fitness function %s' % base._fitness_function)
            else:
                v.fitness = -base._ytr.SSE(v.hy * base._mask)
        else:
            if base._multiple_outputs:
                _ = np.mean([a.SAE(b.mul(c)) for a, b, c in zip(base._ytr, v.hy, base._mask)])
                v.fitness = - _
            else:
                v.fitness = -base._ytr.SAE(v.hy * base._mask)

    def fitness_vs(self, v):
        """Fitness function in the validation set
        In classification it uses BER and RSE in regression"""
        base = self._base
        if base._classifier:
            if base._multiple_outputs:
                v.fitness_vs = v._error
                # if base._fitness_function == 'macro-F1':
                #     v.fitness_vs = v._error
                # elif base._fitness_function == 'BER':
                #     v.fitness_vs = v._error
                # elif base._fitness_function == 'macro-Precision':
                #     v.fitness_vs = v._error
                # elif base._fitness_function == 'accDotMacroF1':
                #     v.fitness_vs = v._error
                # elif base._fitness_function == 'macro-RecallF1':
                #     v.fitness_vs = v._error
                # elif base._fitness_function == 'F1':
                #     v.fitness_vs = v._error
                # else:
                #     v.fitness_vs = - v._error.dot(base._mask_vs) / base._mask_vs.sum()
            else:
                v.fitness_vs = -((base.y - v.hy.sign()).sign().fabs() *
                                 base._mask_vs).sum()
        else:
            mask = base._mask
            y = base.y
            hy = v.hy
            if not isinstance(mask, list):
                mask = [mask]
                y = [y]
                hy = [hy]
            fit = []
            for _mask, _y, _hy in zip(mask, y, hy):
                m = (_mask + -1).fabs()
                x = _y * m
                y = _hy * m
                a = (x - y).sq().sum()
                b = (x + -x.sum() / x.size()).sq().sum()
                fit.append(-a / b)
            v.fitness_vs = np.mean(fit)

    def set_fitness(self, v):
        """Set the fitness to a new node.
        Returns false in case fitness is not finite"""
        base = self._base
        self.fitness(v)
        if not np.isfinite(v.fitness):
            self.del_error(v)
            return False
        if base._tr_fraction < 1:
            self.fitness_vs(v)
            if not np.isfinite(v.fitness_vs):
                self.del_error(v)
                return False
        self.del_error(v)
        return True
        
